@@ -1,7 +1,7 @@
 from warnings import warn
 from optimade.models import StructureResource as OptimadeStructure
+from optimade.models import StructureFeatures
 from optimade.adapters.exceptions import ConversionError
-from optimade.adapters.structures.utils import pad_positions
 
 try:
     from jarvis.core.atoms import Atoms
@@ -28,16 +28,14 @@ def get_jarvis_atoms(optimade_structure: OptimadeStructure) -> Atoms:
     attributes = optimade_structure.attributes
 
     # Cannot handle partial occupancies
-    if "disorder" in attributes.structure_features:
+    if StructureFeatures.DISORDER in attributes.structure_features:
         raise ConversionError(
             "jarvis-tools cannot handle structures with partial occupancies."
         )
 
-    cartesian_site_positions, _ = pad_positions(attributes.cartesian_site_positions)
-
     return Atoms(
         lattice_mat=attributes.lattice_vectors,
         elements=[specie.name for specie in attributes.species],
-        coords=cartesian_site_positions,
+        coords=attributes.cartesian_site_positions,
         cartesian=True,
     )
